import typing
from fastapi import Response
from loguru import logger
import inspect
from functools import wraps


class EventHandler(object):

    """A relatively simple "event handler". Pass in async functions and provide a string, and it will call it

    Attributes:
        callbacks (dict): dict of names --> functions
    """

    def __init__(self):
        self.callbacks = {}

    def on(
        self,
        event: str,
        func: typing.Callable = None,
    ):
        """A wrapper over an async function, registers it in .callbacks.

        Args:
            event (str): Event name
            func (None, optional): function to wrap around

        Returns:
            <function>: returns the wrapped function
        """

        def on(func):
            if not inspect.iscoroutinefunction(func):
                raise TypeError("Function must be a async function.")

            if event not in self.callbacks:
                logger.debug(f"Added {event} to corresponding function to {func}")
                self.callbacks[event] = {
                    "settings": {},
                    "function": func,
                }
            else:
                raise TypeError("Events can only have one corresponding handler.")

            return func

        return on(func) if func else on

    def check_event_exists(self, event: str) -> bool:
        """Checks if the event in ``.callbacks``

        Args:
            event (str): event name

        Returns:
            bool: returns if the event is in callbacks.
        """
        return event in self.callbacks

    def return_event_settings(self, event: str) -> dict:
        if self.check_event_exists(event) == True:
            return self.callbacks[event]["settings"]
        raise TypeError(
            f"Event {event} is not in callbacks. Did you register this event?"
        )

    def return_event_function(self, event: str) -> dict:
        if self.check_event_exists(event) == True:
            return self.callbacks[event]["function"]
        raise TypeError(
            f"Event {event} is not in callbacks. Did you register this event?"
        )

    def view_event_function_return_type(self, event: str) -> dict:
        """Get type hint for event functions

        Args:
            event (str): Event name

        Returns:
            dict: Returns .get_type_hints for event
        """
        return typing.get_type_hints(self.callbacks[event]["function"])

    async def emit(self, event: str, *args, **kwargs):
        """'Emits' an event. It will basically call the function from .callbacks and return the function result

        Args:
            event (str): Event name
            *args: extra arguments to pass
            **kwargs: extra kwargs to pass

        Returns:
            function result: returns the function result

        Raises:
            TypeError: raises if event is not registered.
        """
        if event not in self.callbacks:
            raise TypeError(
                f"event {event} does not have a corresponding handler. Did you register this function/event?"
            )

        _look_up_function = self.return_event_function(event)
        return await _look_up_function(**kwargs)



from pydantic import BaseModel, validator, Extra
import typing
from pydantic.error_wrappers import ValidationError

from pydantic.errors import ArbitraryTypeError


try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal


class CommandTypes:

    """Easy access to command types.

    Attributes:
        BOOLEAN (int): Represents Type 5
        CHANNEL (int): Represents Type 7
        INTEGER (int): Represents Type 4
        ROLE (int): Represents Type 8
        STRING (int): Represents Type 3
        SUB_COMMAND (int): Represents Type 1
        SUB_COMMAND_GROUP (int): Represents Type 2
        USER (int): Represents Type 6
    """

    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8


class CommandChoice(BaseModel):

    """Represents a key-value command choice."""

    name: str
    value: str


class CommandOption(BaseModel):

    """Represents a standard command option (not a subcommand)."""

    class Config:
        arbitrary_types_allowed = True

    name: str
    description: str
    type: int
    required: bool = False
    choices: typing.Optional[
        typing.Union[typing.List[dict], typing.List[CommandChoice]]
    ] = None

    options: typing.Optional[
        typing.Union[typing.List[CommandChoice], typing.List]
    ] = None

    # @validator("options")
    # def options_allow_only_if_subcommand(cls, v):
    #   if cls.type != 1:
    #       raise ValidationError("Type must be 1 in order to have options.")
    #   return v


class SubcommandOption(BaseModel):

    """Represents a subcommand, usually you would put this as an option in a DiscordCommand"""

    class Config:
        arbitrary_types_allowed = True

    name: str
    description: str
    type: Literal[2] = 2
    options: typing.List[CommandOption]

    @validator("options")
    def options_must_contain_type_1(cls, v):  # pylint: disable=no-self-argument
        item: CommandOption
        for item_location, item in enumerate(v):
            if item.type != 1:
                raise ValueError(
                    f"CommandOptions <{item.name}> located <{item_location}> must be have type of 1 due to parent being a subcommand."
                )
        return v


class DiscordCommand(BaseModel):

    """Represents a discord command."""

    name: str
    description: str
    options: typing.List[typing.Union[SubcommandOption, CommandOption]]



from fastapi import FastAPI
import typing
from loguru import logger
import uvicorn
from .server import router, DiscordVerificationMiddleware
from .server import interaction as router_interaction
from .register import RegisterCommands
from .register.models import DiscordCommand
from .models import IncomingApplicationCommand
import aiojobs

from .errors.network import DiscordAPIError


if typing.TYPE_CHECKING:
    import httpx  # pragma: no cover
    from .eventer import EventHandler  # pragma: no cover


class Dispike(object):

    """Dispike - python library for interacting with discord slash commands via an independently hosted server.

    *Powered by FastAPI*
    """

    def __init__(self, client_public_key: str, bot_token: str, application_id: str):
        """Initialize Dispike Object

        Args:
            client_public_key (str): Discord provided client public key.
            bot_token (str): Discord provided bot token. You must create a bot user to view this!
            application_id (str): Discord provided Client ID
        """
        self._bot_token = bot_token
        self._application_id = application_id
        self._registrator = RegisterCommands(
            application_id=self._application_id, bot_token=self._bot_token
        )
        self._internal_application = FastAPI()
        self._internal_application.add_middleware(
            DiscordVerificationMiddleware, client_public_key=client_public_key
        )
        self._internal_application.include_router(router=router)

    def reset_registration(self, new_bot_token=None, new_application_id=None):
        """This method resets the built-in RgeisterCommands.
        You should not have to call this method directly.

        Call it only if you change the client id or bot token.

        Args:
            new_bot_token (None, optional): Description
            new_application_id (None, optional): Description

        Returns:
            TYPE: bool
        """
        try:
            if new_bot_token == None:
                _bot_token = self._bot_token
            else:
                _bot_token = new_bot_token

            if new_application_id == None:
                _application_id = self._application_id
            else:
                _application_id = new_application_id
            self._registrator = RegisterCommands(
                application_id=_application_id, bot_token=_bot_token
            )
            self._bot_token = _bot_token
            self._application_id = _application_id
            return True
        except Exception:
            return False

    @staticmethod
    async def background(function: typing.Callable, *args, **kwargs):
        _schedule = await aiojobs.create_scheduler()
        await _schedule.spawn(function(*args, **kwargs))

    @property
    def interaction(self) -> "EventHandler":
        """Returns an already initialized ``EventHandler`` object.
        You will use this method to handle incoming commands.

        Returns:
            EventHandler: shared EventHandler
        """
        return router_interaction

    @property
    def referenced_application(self) -> FastAPI:
        """Returns the internal FastAPI object that was initialized.
        You are welcome to edit this with the appropriate settings found in
        the FastAPI docs.

        Returns:
            FastAPI: a pre-configured FastAPI object with required middlewares.
        """
        return self._internal_application

    @property
    def register(self) -> RegisterCommands.register:
        """Returns a shortcut the RegisterCommands.register function

        Returns:
            RegisterCommands.register: internal RegisterCommands Object
        """
        return self._registrator.register

    @property
    def shared_client(self) -> "httpx.Client":
        """Returns a pre-initialized ``httpx.Client`` that is used for requests internally.

        Returns:
            httpx.Client: used for network requests to discord.
        """
        return self._registrator._client

    def get_commands(
        self, guild_only=False, guild_id_passed=None
    ) -> typing.List[IncomingApplicationCommand]:
        """Returns a list of ``DiscordCommands`` either globally or for a specific guild.

        Args:
            guild_only (bool, optional): whether to target a guild. Defaults to False.
            guild_id_passed ([type], optional): guild id if guild_only is set to True. Defaults to None.

        Returns:
            typing.List[DiscordCommand]: Array of DiscordCommand

        Raises:
            DiscordAPIError: any Discord returned errors.
        """
        if guild_only == True:
            if guild_id_passed == False or not isinstance(guild_id_passed, str):
                raise TypeError(
                    "You cannot have guild_only == True and NOT pass any guild id."
                )
            _url = f"/guilds/{guild_id_passed}/commands"
        else:
            _url = f"/commands"

        try:
            _send_request = self._registrator._client.get(
                _url, headers=self._registrator.request_headers
            )
            if _send_request.status_code == 200:
                return [IncomingApplicationCommand(**x) for x in _send_request.json()]

            raise DiscordAPIError(_send_request.status_code, _send_request.text)
        except DiscordAPIError:
            logger.exception("Discord API Failure.")
            raise
        except Exception:
            logger.exception("Unknown exception returned")
            raise

    def edit_command(
        self,
        command_id: int,
        new_command: DiscordCommand,
        guild_only=False,
        guild_id_passed=None,
    ) -> DiscordCommand:
        """Edits a command provided with a command_id and a valid new command.

        Args:
            command_id (int): Command ID
            new_command (DiscordCommand): A valid DiscordCommand object (or a dict with proper syntax, if a dict is passed no verification will be made and discord will return the syntax error)
            guild_only (bool, optional): whether to target a guild. Defaults to False.
            guild_id_passed ([type], optional): guild id if guild_only is set to True. Defaults to None.

        Returns:
            DiscordCommand: Returns the DiscordCommand object created. (Will return a DiscordCommand irregardless of new_command)

        Raises:
            TypeError: Invalid types passed.
            DiscordAPIError: any Discord returned errors.
        """
        if not isinstance(new_command, (DiscordCommand, dict)):
            raise TypeError("New command must be a DiscordCommand or a valid dict.")

        if guild_only == True:
            if guild_id_passed == False:
                raise TypeError(
                    "You cannot have guild_only == True and NOT pass any guild id."
                )
            _url = f"/guilds/{guild_id_passed}/commands/{command_id}"
        else:
            _url = f"/commands/{command_id}"

        _new_command = new_command.dict()
        try:
            _send_request = self._registrator._client.patch(
                _url, headers=self._registrator.request_headers, json=_new_command
            )
            if _send_request.status_code != 200:
                raise DiscordAPIError(_send_request.status_code, _send_request.text)
            return DiscordCommand(**_send_request.json())
        except DiscordAPIError:
            logger.exception("Discord API Failure.")
            return False
        except Exception:
            logger.exception("Unknown exception returned")
            return False

    def delete_command(
        self, command_id: int, guild_only=False, guild_id_passed=None
    ) -> bool:
        """Deletes a command, provided with a command_id

        Args:
            command_id (int): Command ID required
            guild_only (bool, optional): Whether to be a global action or target a guild. Defaults to False.
            guild_id_passed ([type], optional): Guild ID if guild_only is set to True. Defaults to None.

        Returns:
            bool: True if status code is 201, otherwise will LOG exception and return False.

        Raises:
            TypeError: Invalid types passed.
            DiscordAPIError: any Discord returned errors.
        """
        if guild_only == True:
            if guild_id_passed == False:
                raise TypeError(
                    "You cannot have guild_only == True and NOT pass any guild id."
                )
            _url = f"/guilds/{guild_id_passed}/commands/{command_id}"
        else:
            _url = f"/commands/{command_id}"

        try:
            _send_request = self._registrator._client.delete(
                _url, headers=self._registrator.request_headers
            )
            if _send_request.status_code != 204:
                raise DiscordAPIError(_send_request.status_code, _send_request.text)
            return True
        except DiscordAPIError:
            logger.exception("Discord API Failure.")
            return False
        except Exception:
            logger.exception("Unknown exception returned")
            return False

    @staticmethod
    def _return_uvicorn_run_function():
        """Import uvicorn, only exists to make testing easier. You do not need to import this.

        Raises:
            SystemExit: If uvicorn is not installed

        Returns:
            uvicorn: If everything works out.
        """
        try:
            import uvicorn

            return uvicorn
        except Exception:
            raise SystemExit(
                "Uvicorn is not installed. Please use a different webserver pointing to <..>.referenced_application"
            )

    def run(self, port: int = 5000):

        """Runs the bot with the already-installed Uvicorn webserver.

        Args:
            port (int, optional): Port to run the bot over. Defaults to 5000.
        """

        uvicorn = self._return_uvicorn_run_function()
        uvicorn.run(app=self.referenced_application, port=port)


from fastapi import APIRouter, Request, Response, BackgroundTasks
from fastapi.responses import PlainTextResponse
from loguru import logger
from .middlewares.verification import DiscordVerificationMiddleware
from .models.incoming import (
    IncomingDiscordInteraction,
    IncomingDiscordOptionList,
    SubcommandIncomingDiscordOptionList,
    IncomingDiscordOption,
)
from .eventer import EventHandler
from .eventer_helpers.determine_event_information import determine_event_information
from .response import DiscordResponse
import json
import typing


router = APIRouter()
interaction = EventHandler()


_RAISE_FOR_TESTING = False


@router.get("/ping")
async def ping():
    return PlainTextResponse(
        "If you see this, Your instance is working and accepting requests."
    )


@router.post("/interactions")
async def handle_interactions(request: Request) -> Response:
    logger.info("interaction recieved.")

    _get_request_body = json.loads(request.state._cached_body.decode())
    logger.info(_get_request_body)
    if _get_request_body["type"] == 1:
        logger.info("handling ACK Ping.")
        return {"type": 1}

    _parse_to_object = IncomingDiscordInteraction(**_get_request_body)
    _event_name, arguments = determine_event_information(_parse_to_object)
    logger.info(f"event name: {_event_name}")
    if interaction.check_event_exists(_event_name) == False:
        logger.debug("discarding event not existing.")
        return {"type": 5}

    _event_settings = interaction.return_event_settings(_event_name)

    arguments["ctx"] = _parse_to_object

    # Check the type hint for the return type, fallback for checking the type if no hints are provided
    try:
        _type_hinted_request = interaction.view_event_function_return_type(_event_name)
        _type_hinted_returned_value = _type_hinted_request["return"]
        if _type_hinted_returned_value == DiscordResponse:
            _get_res = await interaction.emit(_event_name, **arguments)

            logger.debug(_get_res.response)
            return _get_res.response
        elif _type_hinted_returned_value == dict:
            return await interaction.emit(_event_name, **arguments)
    except KeyError:
        logger.error(
            "unable to find return value for type hint.. resorting to guessing.."
        )
        if _RAISE_FOR_TESTING == True:
            raise AssertionError("No hinting!")  # pragma: no cover
    except Exception:
        logger.exception("unhandled exception for returning hinted value")
        raise

    interaction_data = await interaction.emit(_event_name, **arguments)
    if isinstance(interaction_data, DiscordResponse):
        interaction_data: DiscordResponse
        return interaction_data.response

    if isinstance(interaction_data, dict):
        return interaction_data

    # Backup response, simply acknowledge. (Type 5)
    return {"type": 5}



from ..models.incoming import (
    IncomingDiscordInteraction,
    IncomingDiscordOptionList,
    SubcommandIncomingDiscordOptionList,
    IncomingDiscordOption,
)
import typing
from loguru import logger


def determine_event_information(
    interaction: IncomingDiscordInteraction,
) -> typing.Tuple[str, dict]:

    if isinstance(interaction, IncomingDiscordInteraction) == True:
        if (
            isinstance(interaction.data.options[0], SubcommandIncomingDiscordOptionList)
            == True
        ):
            # subcommand event names will be must be xxx.xxx
            _sub_command_arguments = {}
            _event_name = f"{interaction.data.name}.{interaction.data.options[0].name}"
            for argument in interaction.data.options[0].options:
                _sub_command_arguments[argument.name] = argument.value

            return _event_name, _sub_command_arguments

        elif isinstance(interaction.data.options[0], IncomingDiscordOption) == True:
            _command_arguments = {}

            for argument in interaction.data.options:
                argument: IncomingDiscordOption
                _command_arguments[argument.name] = argument.value

            return interaction.data.name, _command_arguments
    raise TypeError("unable to determine eventname.")


try:
    from importlib_metadata import version
except Exception:
    try:
        from importlib.metadata import version
    except Exception:

        def version(*args, **kwargs):
            return "unknown"


try:
    __version__ = version(__package__)
except Exception:
    __version__ = "unknown"

from .main import Dispike


import datetime
import re


def parse_time(timestamp):
    if timestamp:
        return datetime.datetime(
            *map(int, re.split(r"[^\d]", timestamp.replace("+00:00", "")))
        )
    return None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dispike',
 'dispike.errors',
 'dispike.eventer_helpers',
 'dispike.followup',
 'dispike.followup.storage',
 'dispike.helper',
 'dispike.middlewares',
 'dispike.models',
 'dispike.models.discord_types',
 'dispike.register',
 'dispike.register.models']

package_data = \
{'': ['*']}

install_requires = \
['PyNaCl>=1.4.0,<2.0.0',
 'aiojobs>=0.3.0,<0.4.0',
 'async-timeout>=3.0.1,<4.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'httpx>=0.16.1,<0.17.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.7.3,<2.0.0',
 'typing-extensions>=3.7.4,<4.0.0',
 'uvicorn>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'dispike',
    'version': '0.8.0a0',
    'description': 'library for interacting with discord slash commands via an independently hosted server. Powered by FastAPI',
    'long_description': "# dispike\n\n***\n[![codecov](https://codecov.io/gh/ms7m/dispike/branch/master/graph/badge.svg?token=E5AXLZDP9O)](https://codecov.io/gh/ms7m/dispike) ![Test Dispike](https://github.com/ms7m/dispike/workflows/Test%20Dispike/badge.svg?branch=master) ![PyPi Link](https://img.shields.io/badge/Available%20on%20PyPi-Dispike-blue?logo=pypi&link=%22https://pypi.org/project/dispike%22) ![PyPiVersion](https://img.shields.io/badge/dynamic/json?color=blue&label=PyPi%20Version&query=%24.info.version&url=https%3A%2F%2Fpypi.org%2Fpypi%2Fdispike%2Fjson) ![Docs](https://img.shields.io/badge/Docs-Available-lightgrey?link=https://dispike.ms7m.me/) \n\n***\n\n\n\nan *extremely-extremely* early WIP library for easily creating REST-based webhook bots for discord using the new Slash Commands feature. \n\nPowered by FastAPI.\n\n\n***\n\n## Learn more\n- Read documentation [here](https://dispike.ms7m.me)\n- See example bot [here](https://github.com/ms7m/dispike-example-bot)\n\n\n\n\n## Caveats\n\n- ~~Does not handle registring new commands.~~\n- ~~Does not handle anything other then string responses. (However you are free to return any valid dict in your handler.)~~\n- ~~Not on PyPi~~\n- Does not speak over the discord gateway. You'll need a server to handle requests and responses.\n- Python 3.6+\n- Does not support the following endpoints\n  - [Create Followup Message](https://discord.com/developers/docs/interactions/slash-commands#create-followup-message)\n  - [Edit Followup Message](https://discord.com/developers/docs/interactions/slash-commands#edit-followup-message)\n  - [Delete Followup Message](https://discord.com/developers/docs/interactions/slash-commands#delete-followup-message)\n- Handling followup messages.\n\n\n\n# Development\n\nHelp is wanted in mantaining this library. Please try to direct PRs to the ``dev`` branch, and use black formatting (if possible).\n\n![Test Dispike](https://github.com/ms7m/dispike/workflows/Test%20Dispike/badge.svg?branch=dev)\n",
    'author': 'Mustafa Mohamed',
    'author_email': 'ms7mohamed@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ms7m/dispike',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
